import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import torch
import numpy as np

def conv3x3x3(in_planes, out_planes, stride=1, dilation=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        dilation=dilation,
        stride=stride,
        padding=dilation,
        bias=False)


def downsample_basic_block(x, planes, stride, no_cuda=False):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3),
        out.size(4)).zero_()
    if not no_cuda:
        if isinstance(out.data, torch.cuda.FloatTensor):
            zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out

def conv_block(in_channels, out_channels, kernel_size, stride, padding, dilation):
    return nn.Sequential(nn.Conv3d(in_channels, out_channels, kernel_size, stride, padding, dilation),
                         nn.Dropout3d(),
                         nn.InstanceNorm3d(out_channels),
                         nn.LeakyReLU())
    
def encoder_block(in_channels, out_channels, kernel_size = 3, stride = 2, padding = 1, dilation = 1):
    return nn.Sequential(conv_block(in_channels, out_channels, kernel_size, stride, padding = padding, dilation = dilation),
                         conv_block(out_channels, out_channels, kernel_size, stride = 1, padding = padding, dilation = dilation))


def conv_block_2d(in_channels, out_channels, kernel_size, stride, padding, dilation):
    return nn.Sequential(nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation),
                         nn.Dropout2d(),
                         nn.InstanceNorm2d(out_channels),
                         nn.LeakyReLU())

def encoder_block_2d(in_channels, out_channels, kernel_size = 3, stride = 2, padding = 1, dilation = 1):
    return nn.Sequential(conv_block_2d(in_channels, out_channels, kernel_size, stride, padding = padding, dilation = dilation),
                         conv_block_2d(out_channels, out_channels, kernel_size, stride = 1, padding = padding, dilation = dilation))

def aspp_block(in_channels, out_channels):
    return nn.Sequential(ASPP_Block(in_channels, out_channels),
                         nn.Dropout3d(),
                         nn.InstanceNorm3d(out_channels),
                         nn.LeakyReLU())
                    
    
def aspp_encoder(in_channels, out_channels, kernel_size = 3, stride = 2, padding = 1, dilation = 1):
    return nn.Sequential(conv_block(in_channels, out_channels, kernel_size, stride, padding = padding, dilation = dilation),
                         aspp_block(out_channels, out_channels))


class ASPP(nn.Module):
    def __init__(self, in_channels, out_channels, num_classes):
        super(ASPP, self).__init__()
        
        # 1번 branch = 1x1 convolution → BatchNorm → ReLu
        self.conv_1x1_1 = nn.Conv3d(in_channels, out_channels, kernel_size=1)
        self.bn_conv_1x1_1 = nn.BatchNorm3d(out_channels)

        # 2번 branch = 3x3 convolution w/ rate=6 (or 12) → BatchNorm → ReLu
        self.conv_3x3_1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=6, dilation=6)
        self.bn_conv_3x3_1 = nn.BatchNorm3d(out_channels)

        # 3번 branch = 3x3 convolution w/ rate=12 (or 24) → BatchNorm → ReLu
        self.conv_3x3_2 = nn.Conv3d(in_channels, out_channels, kernel_size=3, stride=1, padding=12, dilation=12)
        self.bn_conv_3x3_2 = nn.BatchNorm3d(out_channels)
    
        # 4번 branch = 3x3 convolution w/ rate=18 (or 36) → BatchNorm → ReLu
        self.conv_3x3_3 = nn.Conv3d(in_channels, out_channels, kernel_size=3, stride=1, padding=18, dilation=18)
        self.bn_conv_3x3_3 = nn.BatchNorm3d(out_channels)

        # 5번 branch = AdaptiveAvgPool2d → 1x1 convolution → BatchNorm → ReLu
        self.avg_pool = nn.AdaptiveAvgPool3d(1)
        self.conv_1x1_2 = nn.Conv3d(in_channels, out_channels, kernel_size=1)
        self.bn_conv_1x1_2 = nn.BatchNorm2d(out_channels)
        
        self.conv_1x1_3 = nn.Conv2d(out_channels * 5, out_channels, kernel_size=1) # (1280 = 5*256)
        self.bn_conv_1x1_3 = nn.BatchNorm2d(out_channels)

        self.conv_1x1_4 = nn.Conv2d(out_channels, num_classes, kernel_size=1)

    def forward(self, feature_map):
        # feature map의 shape은 (batch_size, in_channels, height/output_stride, width/output_stride)

        feature_map_h = feature_map.size()[2] # (== h/16)
        feature_map_w = feature_map.size()[3] # (== w/16)

        # 1번 branch = 1x1 convolution → BatchNorm → ReLu
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out_1x1 = F.relu(self.bn_conv_1x1_1(self.conv_1x1_1(feature_map)))
        # 2번 branch = 3x3 convolution w/ rate=6 (or 12) → BatchNorm → ReLu
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out_3x3_1 = F.relu(self.bn_conv_3x3_1(self.conv_3x3_1(feature_map)))
        # 3번 branch = 3x3 convolution w/ rate=12 (or 24) → BatchNorm → ReLu
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out_3x3_2 = F.relu(self.bn_conv_3x3_2(self.conv_3x3_2(feature_map)))
        # 4번 branch = 3x3 convolution w/ rate=18 (or 36) → BatchNorm → ReLu
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out_3x3_3 = F.relu(self.bn_conv_3x3_3(self.conv_3x3_3(feature_map)))

        # 5번 branch = AdaptiveAvgPool2d → 1x1 convolution → BatchNorm → ReLu
        # shape: (batch_size, in_channels, 1, 1)
        out_img = self.avg_pool(feature_map) 
        # shape: (batch_size, out_channels, 1, 1)
        out_img = F.relu(self.bn_conv_1x1_2(self.conv_1x1_2(out_img)))
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out_img = F.upsample(out_img, size=(feature_map_h, feature_map_w), mode="bilinear")

        # shape: (batch_size, out_channels * 5, height/output_stride, width/output_stride)
        out = torch.cat([out_1x1, out_3x3_1, out_3x3_2, out_3x3_3, out_img], 1) 
        # shape: (batch_size, out_channels, height/output_stride, width/output_stride)
        out = F.relu(self.bn_conv_1x1_3(self.conv_1x1_3(out))) 
        # shape: (batch_size, num_classes, height/output_stride, width/output_stride)
        out = self.conv_1x1_4(out) 

        return out

class ASPP_Block(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv1x1 = conv_block(in_channels, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1)
        self.conv3x3_1 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 6, dilation = 6)
        self.conv3x3_2 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 12, dilation = 12)
        self.conv3x3_3 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 18, dilation = 18)
        self.avg_pool = nn.Sequential(nn.AdaptiveAvgPool3d(1),
                        conv_block(in_channels, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1))

        self.concat_conv = conv_block(out_channels*5, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1)

    
    def forward(self, X):
        _, _, w, h, d = X.size()
        X_1x1 = self.conv1x1(X)
        X_3x3_6 = self.conv3x3_1(X)
        X_3x3_12 = self.conv3x3_2(X)
        X_3x3_18 = self.conv3x3_3(X)
        X_avg_pool = self.avg_pool(X)
        X_avg_pool = F.upsample(X_avg_pool, size=(w, h, d), mode="trilinear")
        out = torch.cat([X_1x1, X_3x3_6, X_3x3_12, X_3x3_18, X_avg_pool], 1)
        out = self.concat_conv(out)

        return out
        

def conv3x3x3(in_planes, out_planes, stride=1, dilation=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        dilation=dilation,
        stride=stride,
        padding=dilation,
        bias=False)


def downsample_basic_block(x, planes, stride, no_cuda=False):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3),
        out.size(4)).zero_()
    if not no_cuda:
        if isinstance(out.data, torch.cuda.FloatTensor):
            zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3x3(inplanes, planes, stride=stride, dilation=dilation)
        self.bn1 = nn.BatchNorm3d(planes)
        self.relu = nn.LeakyReLU(inplace = True)
        self.conv2 = conv3x3x3(planes, planes, dilation=dilation)
        self.bn2 = nn.BatchNorm3d(planes)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv3d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm3d(planes)
        self.conv2 = nn.Conv3d(
            planes, planes, kernel_size=3, stride=stride, dilation=dilation, padding=dilation, bias=False)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes * 4)
        self.relu = nn.LeakyReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self,
                 block,
                 layers,
                 sample_input_D,
                 sample_input_H,
                 sample_input_W,
                 num_seg_classes,
                 shortcut_type='B',
                 no_cuda = False):
        self.inplanes = 64
        self.no_cuda = no_cuda
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv3d(
            1,
            64,
            kernel_size=7,
            stride=(2, 2, 2),
            padding=(3, 3, 3),
            bias=False)
            
        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.LeakyReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], shortcut_type)
        self.layer2 = self._make_layer(
            block, 128, layers[1], shortcut_type, stride=2)
        self.layer3 = self._make_layer(
            block, 256, layers[2], shortcut_type, stride=1, dilation=2)
        self.layer4 = self._make_layer(
            block, 512, layers[3], shortcut_type, stride=1, dilation=4)

        self.conv_seg = nn.Sequential(
                                        nn.ConvTranspose3d(
                                        512 * block.expansion,
                                        32,
                                        2,
                                        stride=2
                                        ),
                                        nn.BatchNorm3d(32),
                                        nn.LeakyReLU(inplace=True),
                                        nn.Conv3d(
                                        32,
                                        32,
                                        kernel_size=3,
                                        stride=(1, 1, 1),
                                        padding=(1, 1, 1),
                                        bias=False), 
                                        nn.BatchNorm3d(32),
                                        nn.LeakyReLU(inplace=True),
                                        nn.Conv3d(
                                        32,
                                        num_seg_classes,
                                        kernel_size=1,
                                        stride=(1, 1, 1),
                                        bias=False) 
                                        )

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                m.weight = nn.init.kaiming_normal(m.weight, mode='fan_out')

    def _make_layer(self, block, planes, blocks, shortcut_type, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if shortcut_type == 'A':
                downsample = partial(
                    downsample_basic_block,
                    planes=planes * block.expansion,
                    stride=stride,
                    no_cuda=self.no_cuda)
            else:
                downsample = nn.Sequential(
                    nn.Conv3d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False), nn.BatchNorm3d(planes * block.expansion))

        layers = []
        layers.append(block(self.inplanes, planes, stride=stride, dilation=dilation, downsample=downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.conv_seg(x)

        return x


def resnet10(**kwargs):
    """Constructs a ResNet-10 model.
    """
    model = ResNet(BasicBlock, [1, 1, 1, 1], **kwargs)
    return model


def resnet18(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    return model


def resnet34(**kwargs):
    """Constructs a ResNet-34 model.
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    return model


def resnet50(**kwargs):
    """Constructs a ResNet-50 model.
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    return model


def resnet152(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    return model


def resnet200(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 24, 36, 3], **kwargs)
    return model



class ASPP_Block(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv1x1 = conv_block(in_channels, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1)
        self.conv3x3_1 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 6, dilation = 6)
        self.conv3x3_2 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 12, dilation = 12)
        self.conv3x3_3 = conv_block(in_channels, out_channels, kernel_size = 3, stride = 1, padding = 18, dilation = 18)
        self.avg_pool = nn.Sequential(nn.AvgPool3d(1),
                        conv_block(in_channels, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1))

        self.concat_conv = conv_block(out_channels*5, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1)

    
    def forward(self, X):
        _, _, w, h, d = X.size()
        X_1x1 = self.conv1x1(X)
        X_3x3_6 = self.conv3x3_1(X)
        X_3x3_12 = self.conv3x3_2(X)
        X_3x3_18 = self.conv3x3_3(X)
        X_avg_pool = self.avg_pool(X)
        X_avg_pool = F.upsample(X_avg_pool, size=(w, h, d), mode="trilinear")
        out = torch.cat([X_1x1, X_3x3_6, X_3x3_12, X_3x3_18, X_avg_pool], 1)
        out = self.concat_conv(out)

        return out


class My_ASPP(nn.Module):
    def __init__(self, in_channels, out_channels, stride = 1):
        super().__init__()
        self.conv1x1 = conv_block(in_channels//5, out_channels//5, kernel_size = 1, stride = 1, padding = 0, dilation = 1)
        self.conv3x3_0 = conv_block(in_channels//5, out_channels//5, kernel_size = 3, stride = 1, padding = 1, dilation = 1)
        self.conv3x3_1 = conv_block(in_channels//5, out_channels//5, kernel_size = 3, stride = 1, padding = 6, dilation = 6)
        self.conv3x3_2 = conv_block(in_channels//5, out_channels//5, kernel_size = 3, stride = 1, padding = 12, dilation = 12)
        self.conv3x3_3 = conv_block(in_channels//5, out_channels//5, kernel_size = 3, stride = 1, padding = 18, dilation = 18)
        self.conv3x3_4 = conv_block(in_channels, out_channels, kernel_size = 3, stride = stride, padding = 1, dilation = 1)

        

    
    def forward(self, X):
        _, _, w, h, d = X.size()
        X_1x1 = self.conv1x1(X)
        X_3x3_1 = self.conv3x3_0(X)
        X_3x3_6 = self.conv3x3_1(X)
        X_3x3_12 = self.conv3x3_2(X)
        X_3x3_18 = self.conv3x3_3(X)
        return self.conv3x3_4(torch.cat((X_1x1, X_3x3_1, X_3x3_6, X_3x3_12, X_3x3_18), 1))



def myaspp_block(in_channels, out_channels):
    return nn.Sequential(My_ASPP(in_channels, out_channels),
                         nn.Dropout3d(),
                         nn.InstanceNorm3d(out_channels),
                         nn.LeakyReLU())
def myaspp_encoder(in_channels, out_channels, stride):
    return nn.Sequential(myaspp_block(in_channels, out_channels, stride = 1),
                         myaspp_block(in_channels, out_channels, stride = stride))


class ConvGRUCell(nn.Module):
    def __init__(self, input_size, input_dim, hidden_dim, kernel_size, bias = 0, dtype = float):
        """
        Initialize the ConvLSTM cell
        :param input_size: (int, int)
            Height and width of input tensor as (height, width).
        :param input_dim: int
            Number of channels of input tensor.
        :param hidden_dim: int
            Number of channels of hidden state.
        :param kernel_size: (int, int)
            Size of the convolutional kernel.
        :param bias: bool
            Whether or not to add the bias.
        :param dtype: torch.cuda.FloatTensor or torch.FloatTensor
            Whether or not to use cuda.
        """
        super(ConvGRUCell, self).__init__()
        self.height, self.width = input_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.hidden_dim = hidden_dim
        self.bias = bias
        self.dtype = dtype

        self.conv_gates = nn.Conv2d(in_channels=input_dim + hidden_dim,
                                    out_channels=2*self.hidden_dim,  # for update_gate,reset_gate respectively
                                    kernel_size=kernel_size,
                                    padding=self.padding,
                                    bias=self.bias)

        self.conv_can = nn.Conv2d(in_channels=input_dim+hidden_dim,
                              out_channels=self.hidden_dim, # for candidate neural memory
                              kernel_size=kernel_size,
                              padding=self.padding,
                              bias=self.bias)


    def forward(self, input_tensor, h_cur):
        """
        :param self:
        :param input_tensor: (b, c, h, w)
            input is actually the target_model
        :param h_cur: (b, c_hidden, h, w)
            current hidden and cell states respectively
        :return: h_next,
            next hidden state
        """
        combined = torch.cat([input_tensor, h_cur], dim=1)
        combined_conv = self.conv_gates(combined)

        gamma, beta = torch.split(combined_conv, self.hidden_dim, dim=1)
        reset_gate = torch.sigmoid(gamma)
        update_gate = torch.sigmoid(beta)

        combined = torch.cat([input_tensor, reset_gate*h_cur], dim=1)
        cc_cnm = self.conv_can(combined)
        cnm = torch.tanh(cc_cnm)

        h_next = (1 - update_gate) * h_cur + update_gate * cnm
        return h_next


import os
import numpy as np

def get_imbalance_weight(root_dir, n_classes):
    file_list = os.listdir(root_dir)
    result = np.zeros(n_classes)
    for file in file_list:
        if 'npz' in file:
            seg = np.load(os.path.join(root_dir, file))['y']
            id, num = np.unique(seg, return_counts = True)
            result[id.astype(int)] += num
    
    
    return (result.sum()/result).astype(float)



def get_maximum_patch_size(root_dir, n_classes):
    file_list = os.listdir(root_dir)
    result = np.array((0,0,0))
    npz_list = []
    pkl_list = []
    
    for file in file_list:
        if 'npz' in file:
            npz_list.append(file)
        
        elif 'pkl' in file:
            pkl_list.append(file)
        
    
    npz_list = sorted(npz_list)
    pkl_list = sorted(pkl_list)

    for i in range(len(npz_list)):
        seg = np.load(os.path.join(root_dir, file))['y']
        result = np.max((result,np.array(seg.shape)), axis = 0)
    
    return result


from typing import Iterable

import torch
from torch.optim._multi_tensor import SGD

__all__ = ["SAMSGD"]


class SAMSGD(SGD):
    """ SGD wrapped with Sharp-Aware Minimization
    Args:
        params: tensors to be optimized
        lr: learning rate
        momentum: momentum factor
        dampening: damping factor
        weight_decay: weight decay factor
        nesterov: enables Nesterov momentum
        rho: neighborhood size
    """

    def __init__(self,
                 params: Iterable[torch.Tensor],
                 lr: float,
                 momentum: float = 0,
                 dampening: float = 0,
                 weight_decay: float = 0,
                 nesterov: bool = False,
                 rho: float = 0.05,
                 ):
        if rho <= 0:
            raise ValueError(f"Invalid neighborhood size: {rho}")
        super().__init__(params, lr, momentum, dampening, weight_decay, nesterov)
        # todo: generalize this
        if len(self.param_groups) > 1:
            raise ValueError("Not supported")
        self.param_groups[0]["rho"] = rho

    @torch.no_grad()
    def step(self,
             closure
             ) -> torch.Tensor:
        """
        Args:
            closure: A closure that reevaluates the model and returns the loss.
        Returns: the loss value evaluated on the original point
        """
        closure = torch.enable_grad()(closure)
        loss = closure().detach()

        for group in self.param_groups:
            grads = []
            params_with_grads = []

            rho = group['rho']
            # update internal_optim's learning rate

            for p in group['params']:
                if p.grad is not None:
                    # without clone().detach(), p.grad will be zeroed by closure()
                    grads.append(p.grad.clone().detach())
                    params_with_grads.append(p)
            device = grads[0].device

            # compute \hat{\epsilon}=\rho/\norm{g}\|g\|
            grad_norm = torch.stack([g.detach().norm(2).to(device) for g in grads]).norm(2)
            epsilon = grads  # alias for readability
            torch._foreach_mul_(epsilon, rho / grad_norm)

            # virtual step toward \epsilon
            torch._foreach_add_(params_with_grads, epsilon)
            # compute g=\nabla_w L_B(w)|_{w+\hat{\epsilon}}
            closure()
            # virtual step back to the original point
            torch._foreach_sub_(params_with_grads, epsilon)

        super().step()
        return loss


import matplotlib.pyplot as plt
import torch

def deep_supervision_visualization(output, target):
    # output shape = (batch size, deepsupervision size, channel size, x, y, z)
    ds = output.shape[1]
    n_cls = output.shape[2]
    center = output.shape[3] // 2
    for d_i in range(ds):
        print("deepsupervision layer", d_i + 1)
        plt.figure(figsize=(3 * (ds), 3*(n_cls+2)))
        for c_i in range(n_cls):
            plt.subplot(int(str(d_i + 1)+str(n_cls+2)+str(c_i+1)))
            plt.imshow(output[0][d_i][c_i][center].detach().cpu(), cmap = 'gray')
            plt.xlabel("class: " + str(c_i))
        plt.subplot(int(str(d_i + 1)+str(n_cls+2)+str(n_cls + 1)))
        plt.imshow(target[0][center].detach().cpu(), vmin = 0, vmax = n_cls)
        plt.xlabel("ground_truth")
        plt.subplot(int(str(d_i + 1)+str(n_cls+2)+str(n_cls + 2)))
        plt.imshow(torch.argmax(output[0][d_i], dim = 0)[center].detach().cpu(), vmin = 0, vmax = n_cls)
        plt.xlabel("predicted")
        plt.show()


def visualization(output, target):
    # output shape = (batch size, channel size, x, y, z)
    n_cls = output.shape[1]
    center = output.shape[2]//2
    plt.figure(figsize=(10,12))

    for c_i in range(n_cls):
        plt.subplot(int(str(1)+str(n_cls+2)+str(c_i+1)))
        plt.imshow(output[0][c_i][center].detach().cpu(), cmap = 'gray')
        plt.xlabel("class: " + str(c_i))
    plt.subplot(int(str(1)+str(n_cls+2)+str(n_cls + 1)))
    plt.imshow(target[0][center].detach().cpu(), vmin = 0, vmax = n_cls)
    plt.xlabel("ground_truth")
    plt.subplot(int(str(1)+str(n_cls+2)+str(n_cls + 2)))
    plt.imshow(torch.argmax(output[0], dim = 0)[center].detach().cpu(), vmin = 0, vmax = n_cls)
    plt.xlabel("predicted")
    plt.show()


import torch
import torch.nn.functional as F
import numpy as np
import math


class Batch_Affine_3D(object):
    def __init__(self, degree = [30, 30, 30], axis = [0,1,2], scale = [0.7, 1.4], use_gpu = True, device = None, is_random = True, pad_value = -10):
        self.use_gpu = use_gpu
        self.transform_matrices = []
        self.axis = axis
        self.degree = np.array((math.radians(degree[0]), math.radians(degree[1]), math.radians(degree[2])))
        self.scale = np.array(scale)
        self.device = device
        self.is_random = is_random
        self.pad_value = pad_value

    def get_matrices_and_coords(self, images):
        if len(images.shape) == 5:
            seg = False
        else:
            seg = True
        if not seg:
            bs, n_modalities, nx, ny, nz = images.shape
        else:
            bs, nx, ny, nz = images.shape
        img_size = [nx,ny,nz] #patch size이기 때문에 동일
        self.coords = self.create_coords(img_size) #Coordinates 생성
        self.matrices = self.get_matrices()

    
    def __call__(self, images):
        if len(images.shape) == 5:
            seg = False
        else:
            seg = True
        if not seg:
            bs, n_modalities, nx, ny, nz = images.shape
        else:
            bs, nx, ny, nz = images.shape
        img_size = [nx,ny,nz] #patch size이기 때문에 동일
        patch_size = np.array(img_size)

        for batch_index in range(bs):
            transformed_coords = self.gpu_calculate_coornidates(self.coords, self.matrices, img_size).long() #coorniates를 Affine 변환
            padding_tuple = (torch.where(transformed_coords == -1)[-3].long(), torch.where(transformed_coords == -1)[-2].long(), torch.where(transformed_coords == -1)[-1].long())
            if not seg:
                images[batch_index, :, :, :, :] = images[batch_index, :, transformed_coords[0], transformed_coords[1], transformed_coords[2]]
                images = images.reshape(bs, n_modalities, nx, ny, nz)
                for nm in range(n_modalities):
                    images[batch_index][nm][padding_tuple] = self.pad_value
            else:
                images[batch_index, :, :, :] = images[batch_index, transformed_coords[0], transformed_coords[1], transformed_coords[2]]
                images = images.reshape(bs, nx, ny, nz)
                images[batch_index][padding_tuple] = 0
            
        return images


    def set_rotate_matrix(self, axis, degree):
        function_list = [self.rotation_x_3d, self.rotation_y_3d, self.rotation_z_3d]
        return function_list[axis](degree)

    def set_scale_matrix(self, scale):
        return torch.eye(3) * 1/scale

    def rotation_x_3d(self, angle):
        return torch.Tensor([[1, 0, 0], [0, math.cos(angle), -math.sin(angle)], [0, math.sin(angle), math.cos(angle)]])

    def rotation_y_3d(self, angle):
        return torch.Tensor([[math.cos(angle), 0, math.sin(angle)], [0, 1, 0], [-math.sin(angle), 0, math.cos(angle)]])

    def rotation_z_3d(self, angle):
        return torch.Tensor([[math.cos(angle), -math.sin(angle), 0], [math.sin(angle), math.cos(angle), 0], [0, 0, 1]])

    def create_coords(self, img_shape):
        tmp = tuple([torch.arange(i) for i in img_shape])
        grid_x, grid_y, grid_z = torch.meshgrid(*tmp)

        return torch.vstack((grid_x.unsqueeze_(0), grid_y.unsqueeze_(0), grid_z.unsqueeze_(0)))
    
    def set_random_scale_degree(self):
        temp = np.random.uniform(0, 1)
        if temp < 0.4:
            degree = np.random.uniform( size = self.degree.shape) * self.degree
            degree = (np.random.uniform(size = 3) < 0.5).astype(int) * degree
            scale = 1
            
        elif temp < 0.8:
            scale = np.random.uniform(low = self.scale[0], high = self.scale[1], size = 1)[0]
            degree = np.array([0,0,0])
        else:
            scale = np.random.uniform(low = self.scale[0], high = self.scale[1], size = 1)[0]
            degree = np.random.uniform( size = self.degree.shape) * self.degree
            degree = (np.random.uniform(size = 3) < 0.5).astype(int) * degree
            
        return scale, degree
    def get_matrices(self):
        transform_matrices = []
        scale, degree = self.set_random_scale_degree()
        for a, d in zip(self.axis, degree):
            transform_matrices.append(self.set_rotate_matrix(a, d))#Rotation Matrix 생성

        if isinstance(scale, np.ndarray):
            transform_matrices.append(self.set_scale_matrix(scale))

        return torch.stack(transform_matrices)

    def gpu_calculate_coornidates(self, coords, matrices, img_size):
        before_shape = coords.shape
        coords = coords.to(self.device).double()
        coords_bound = np.array(img_size).reshape(-1,1)-1 #원본 이미지의 최소, 최대 좌표

        half_img_size = ((torch.Tensor(img_size).to(self.device) - 1)/2).reshape(3,1,1,1)
        coords = coords - half_img_size
        matrices = matrices.to(self.device).double()

        for i in range(len(matrices)-1):
            matrices[i+1] = torch.matmul(matrices[i], matrices[i+1])
    
        coords = torch.matmul(matrices[-1], coords.reshape(3, -1)).reshape(*before_shape)
        coords = coords + half_img_size
        coords = coords.reshape(3, -1).round() #Mapping 해주는 함수
        
        coords[coords < 0 ] = -1
        coords[-3][coords[-3] > img_size[-3] - 2] = -1
        coords[-2][coords[-2] > img_size[-2] - 2] = -1
        coords[-1][coords[-1] > img_size[-1] - 2] = -1
        #coords = torch.min(torch.max(coords,  torch.Tensor([[0], [0], [0]]).long().to(self.device)) ,torch.Tensor(coords_bound).long().to(self.device)).long()
        
        return coords.reshape(*before_shape)


class Batch_Gaussian_Noise(object):
    def __init__(self, device = 0, prob_per_modalities = 0.5):
        self.device = device
        self.prob_per_modalities = prob_per_modalities
    
    def __call__(self, images):
        if len(images.shape) == 4:
            return images

        bs, n_modalities, nx, ny, nz = images.shape
        do_operation = np.random.uniform(0, 1, size = n_modalities) < self.prob_per_modalities

        for i in range(n_modalities):
            if i == 0:
                if do_operation[i]:
                    noise_array = torch.randn((nx,ny,nz)).unsqueeze(0)
                else:
                    noise_array = torch.zeros((nx,ny,nz)).unsqueeze(0)
                continue
            if do_operation[i]:
                noise_array = torch.vstack((noise_array, torch.randn((nx,ny,nz)).unsqueeze(0)))
            else:
                noise_array = torch.vstack((noise_array, torch.zeros((nx,ny,nz)).unsqueeze(0)))
        noise_array = torch.tile(noise_array, (bs, 1,1,1,1)).to(self.device)

        return images + noise_array


class Batch_Gaussian_Blur_3D(object):
    def __init__(self, sigma = 1.2, width = 3, prob_per_modality = 0.5, device = 0):
        array = np.random.uniform(size = 4)
        self.device = device
        gaussian_filter = self.get_kernel(np.random.uniform(1.0, sigma), width)
        self.prob_per_modality = prob_per_modality
        self.weight = gaussian_filter.unsqueeze(0).unsqueeze(0).to(device)

    def __call__(self, images):
        #각 모달리티 별로 다른 연산을 지원한다.

        batch_size = images.shape[0]
        modality_size = images.shape[1]
        do_per_modality = np.random.uniform(0,1, size = modality_size)

        if len(images.shape) == 4:
            return images
    
        result = []
        images = images.permute(1, 0, 2, 3, 4).unsqueeze(1) #n_modalities, batch_size

        for i in range(modality_size):
            if do_per_modality[i] > self.prob_per_modality:
                continue
            images[i] = F.conv3d(images[i], weight = torch.tile(self.weight, (batch_size, 1, 1, 1, 1)).double(), stride = 1, padding = 1, groups = batch_size)
        

        images = images.squeeze(1).permute(1,0,2,3,4)
    
        return images


    def get_kernel(self, sigma, width):
        kernel_shape = (width, width, width)
        tmp = tuple([torch.arange(i) for i in kernel_shape])
        gx, gy, gz = torch.meshgrid(*tmp)
        gx = gx - 1
        gy = gy - 1
        gz = gz - 1
        kernel = torch.exp(-((gx**2 + gy**2 + gz**2)/(2*(sigma**2))))

        return (kernel / kernel.sum())


class Batch_Brightness(object):
    def __init__(self, device = None, rng = [0.7,1.3]):
        self.device = device
        self.rng = rng
    
    def __call__(self, images):
        x = np.random.uniform(self.rng[0], self.rng[1])

        return images * x



class Batch_GammaTransform(object):
    def __init__(self, gamma_range = (0.5, 2), epsilon = 1e-7, device = None, retain_stats = True):
        self.gamma_range = gamma_range
        self.device = device
        self.epsilon = epsilon
        self.retain_stats = retain_stats

    def __call__(self, images):
        if len(images.shape) == 4:
            return images
        bs, n_modalities, nx, ny, nz = images.shape
        
        if np.random.random() < 0.5 and self.gamma_range[0] < 1:
            gamma = torch.FloatTensor(bs, n_modalities,1,1,1).uniform_(self.gamma_range[0], 1).to(self.device)
        else:
            gamma = torch.FloatTensor(bs, n_modalities,1,1,1).uniform_(max(self.gamma_range[0], 1), self.gamma_range[1]).to(self.device)
        
        if self.retain_stats:
            before_mmean = torch.mean(images.view(bs, n_modalities, -1), 2).view(bs, n_modalities, 1, 1, 1)
            before_mstd = torch.std(images.view(bs, n_modalities, -1), 2).view(bs, n_modalities, 1, 1, 1)
        mmax = torch.max(images.view(bs, n_modalities, -1), dim = 2)[0].view(bs, n_modalities, 1, 1, 1)
        mmin = torch.min(images.view(bs, n_modalities, -1), dim = 2)[0].view(bs, n_modalities, 1, 1, 1)
        
        mrange = mmax - mmin 
        temp = (mrange + self.epsilon)
        
        images = torch.pow(((images - mmin) / temp), gamma) * temp + mmin
        

        if self.retain_stats:
            after_mmean = torch.mean(images.view(bs, n_modalities, -1), 2).view(bs, n_modalities, 1, 1, 1)
            after_mstd = torch.std(images.view(bs, n_modalities, -1), 2).view(bs, n_modalities, 1, 1, 1)
            images = images - after_mmean
            images = images / (after_mstd + 1e-8) * before_mstd
            images = images + before_mmean
        
        return images

class Batch_Contrast(object):
    def __init__(self, device = None, contrast_range = [0.75, 1.25], preserve_range = True):
        self.device = device
        self.preserve_range = preserve_range
        self.contrast_range = contrast_range
    
    def __call__(self, images):
        if len(images.shape) == 4:
            return images
        bs, n_modalities, nx, ny, nz = images.shape
    
        if np.random.random() < 0.5 and self.contrast_range[0] < 1:
            factor = torch.FloatTensor(bs, n_modalities,1,1,1).uniform_(self.contrast_range[0], 1).to(self.device)
        else:
            factor = torch.FloatTensor(bs, n_modalities,1,1,1).uniform_(max(self.contrast_range[0], 1), self.contrast_range[1]).to(self.device)

    
        mmean = torch.mean(images.view(n_modalities, -1), 1).view(n_modalities, 1, 1, 1)
        mmax = torch.max(images.view(n_modalities, -1), dim = 1)[0].view(n_modalities, 1, 1, 1)
        mmin = torch.min(images.view(n_modalities, -1), dim = 1)[0].view(n_modalities, 1, 1, 1)
        images = (images - mmean ) * factor + mmean
        if self.preserve_range:
            images = torch.where(images > mmax, mmax, images)
            images = torch.where(images < mmin, mmin, images)

        return images

    
class Batch_Mirroring(object):
    def __init__(self, x_prob = 0.5, y_prob = 0.5, z_prob = 0.5):
        self.x_prob = x_prob
        self.y_prob = y_prob
        self.z_prob = z_prob

    def get_random_variables(self):
        do_mirror = np.random.uniform(0,1, size = 3)
        max_value = np.max(do_mirror)
        self.mirror_axis = []
        if do_mirror[0] == max_value:
            self.mirror_axis.append(-3)
        
        if do_mirror[1] == max_value:
            self.mirror_axis.append(-2)
        
        if do_mirror[2] == max_value:
            self.mirror_axis.append(-1)


    def __call__(self, images):
        if self.mirror_axis:
            return torch.flip(images, self.mirror_axis)
        
        else:
            return images

    
class Batch_Random_Permute(object):
    def __init__(self, permute_index = np.array([-1,-2,-3])):
        self.permute_index = permute_index
    
    def get_random_variables(self):
        np.random.shuffle(self.permute_index)

    def __call__(self, images):
        if len(images.shape) == 5:
            images = images.permute(0, 1, self.permute_index[0], self.permute_index[1], self.permute_index[2])
        
        elif len(images.shape) == 4:
            images = images.permute(0, self.permute_index[0], self.permute_index[1], self.permute_index[2])

        return images

class Batch_Low_Resolution(object):
    def __init__(self, prob = 0.5):
        self.prob_per_modality = 0.5

    def __call__(self, images):
        modality_size = images.shape[1]
        do_per_modality = np.random.uniform(0,1, size = modality_size)
        for i in range(modality_size):
            if do_per_modality[i] > self.prob_per_modality:
                continue
            images[:, i] = F.interpolate(F.interpolate(images[:, i].unsqueeze(0), scale_factor = 0.5, mode ='trilinear', recompute_scale_factor=True, align_corners = True), scale_factor = 2.0, mode ='trilinear',recompute_scale_factor=True, align_corners = True).squeeze(0)
        
        return images


import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torch.autograd import Variable
import math
from functools import partial


class ConvDropoutNormNonlin(nn.Module):
    def __init__(self, input_channels = 4, output_channels = 1,
                 conv_op=nn.Conv3d, conv_kwargs=None,
                 norm_op=nn.InstanceNorm3d, norm_op_kwargs=None,
                 dropout_op=nn.Dropout3d, dropout_op_kwargs=None,
                 nonlin=nn.LeakyReLU, nonlin_kwargs=None):
        super().__init__()
        if nonlin_kwargs is None:
            nonlin_kwargs = {'negative_slope': 1e-2, 'inplace': True}
        if dropout_op_kwargs is None:
            dropout_op_kwargs = {'p': 0.5, 'inplace': True}
        if norm_op_kwargs is None:
            norm_op_kwargs = {'eps': 1e-5, 'affine': True, 'momentum': 0.1}
        if conv_kwargs is None:
            conv_kwargs = {'kernel_size': 3, 'stride': 1, 'padding': 1, 'dilation': 1, 'bias': True}

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        self.conv = self.conv_op(input_channels, output_channels, **self.conv_kwargs)
        if self.dropout_op is not None and self.dropout_op_kwargs['p'] is not None and self.dropout_op_kwargs[
            'p'] > 0:
            self.dropout = self.dropout_op(**self.dropout_op_kwargs)
        else:
            self.dropout = None
        self.instnorm = self.norm_op(output_channels, **self.norm_op_kwargs)
        self.lrelu = self.nonlin(**self.nonlin_kwargs)

    def forward(self, x):
        x = self.conv(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.lrelu(self.instnorm(x))


class DeepLabV3_CT(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.deeplab = torch.hub.load('pytorch/vision:v0.10.0', 'deeplabv3_resnet50', pretrained=True)
        self.deeplab.backbone.conv1 = nn.Conv2d(in_channels, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.deeplab.aux_classifier[-1] = nn.Conv2d(512, out_channels, kernel_size=(1,1), stride=(1,1))
        

    def forward(self, input, hidden):
        
        input = self.deeplab.backbone(input)
        input = self.deeplab.classifier(input)



class DeepSuperVisionUNet2dot5(nn.Module):
    def __init__(self, n_modalities = 1, depth = 256, fianl_output_channels = 2, patch_size = (256,512,512)):
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        encoder_channel1 = 64
        encoder_channel2 = encoder_channel1 * 2
        encoder_channel3 = encoder_channel2 * 2
        encoder_channel4 = encoder_channel3 * 2
        encoder_channel5 = encoder_channel4 * 2

        self.encoder1 = nn.Sequential(
            ConvDropoutNormNonlin(n_modalities * depth, encoder_channel1, conv_kwargs= conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d),
            ConvDropoutNormNonlin(encoder_channel1, encoder_channel1, conv_kwargs = conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d)
        )
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 2}
        conv_kwargs2 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        self.encoder2 = nn.Sequential(
            ConvDropoutNormNonlin(encoder_channel1, encoder_channel2, conv_kwargs= conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d),
            ConvDropoutNormNonlin(encoder_channel2, encoder_channel2, conv_kwargs= conv_kwargs2, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d)
        )
        self.encoder3 = nn.Sequential(
            ConvDropoutNormNonlin(encoder_channel2, encoder_channel3, conv_kwargs= conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d),
            ConvDropoutNormNonlin(encoder_channel3, encoder_channel3, conv_kwargs= conv_kwargs2, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d)
        )
        self.encoder4 = nn.Sequential(
            ConvDropoutNormNonlin(encoder_channel3, encoder_channel4, conv_kwargs= conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d),
            ConvDropoutNormNonlin(encoder_channel4, encoder_channel4, conv_kwargs= conv_kwargs2, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d)
        )
        self.encoder5 = nn.Sequential(
            ConvDropoutNormNonlin(encoder_channel4, encoder_channel5, conv_kwargs= conv_kwargs1, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d),
            ConvDropoutNormNonlin(encoder_channel5, encoder_channel5, conv_kwargs= conv_kwargs2, conv_op = nn.Conv2d, norm_op = nn.InstanceNorm2d, dropout_op = nn.Dropout2d)
        )
    


class Deep_Supervision_UNet3d_Narrow(nn.Module):
    def __init__(self, n_modalities = 4, final_output_channels = 5, patch_size = (128, 128, 128), is_binary = False):
        super().__init__()
        self.patch_size = np.array(patch_size)
        self.networks_depth = self.get_networks_depth()
        self.encoder_blocks = []
        self.decoder_blocks = []
        self.upsample_operations = []
        self.final_output_channels = final_output_channels
        self.is_binary = is_binary
        
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        self.encoder1 = nn.Sequential(
            ConvDropoutNormNonlin(n_modalities, 32, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(32, 32, conv_kwargs = conv_kwargs1, conv_op = nn.Conv3d)
        )
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 2}
        conv_kwargs2 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        self.encoder2 = nn.Sequential(
            ConvDropoutNormNonlin(32, 64, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(64, 64, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder3 = nn.Sequential(
            ConvDropoutNormNonlin(64, 128, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(128, 128, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder4 = nn.Sequential(
            ConvDropoutNormNonlin(128, 256, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(256, 256, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder5 = nn.Sequential(
            ConvDropoutNormNonlin(256, 320, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(320, 320, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )

        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1,  'stride' : 1}
        conv_kwargs2 = {'kernel_size' : 3, 'stride' : 1, 'padding':1}
        transp_kwargs = {'kernel_size' : 2, 'stride' : 1, 'stride' : 2}


        self.decoder1_trans = ConvDropoutNormNonlin(320, 256, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d)
        self.decoder1_conv = nn.Sequential(
            ConvDropoutNormNonlin(256*2, 256, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(256, 256, conv_kwargs = conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder2_trans = ConvDropoutNormNonlin(256, 128, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder2_conv = nn.Sequential(
            ConvDropoutNormNonlin(128*2, 128 , conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(128, 128, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder3_trans = ConvDropoutNormNonlin(128, 64, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder3_conv = nn.Sequential(
            ConvDropoutNormNonlin(64*2, 64, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(64, 64, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder4_trans = ConvDropoutNormNonlin(64, 32, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder4_conv = nn.Sequential(
            ConvDropoutNormNonlin(32*2, 32, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(32, 32, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )


        self.final_layer1 = nn.Sequential(
            nn.Conv3d(32, final_output_channels, 3, 1, 1)
            )
        self.final_layer2 = nn.Sequential(
            nn.Conv3d(64, final_output_channels, 3, 1, 1)
            )
        self.final_layer3 = nn.Sequential(
            nn.Conv3d(128, final_output_channels, 3, 1, 1)
            )
        self.final_layer4 = nn.Sequential(
            nn.Conv3d(256, final_output_channels, 3, 1, 1)
            )

        self.upsample4 = nn.Upsample(scale_factor=8, mode = 'trilinear')
        self.upsample3 = nn.Upsample(scale_factor=4, mode = 'trilinear')
        self.upsample2 = nn.Upsample(scale_factor=2, mode = 'trilinear')
        
    
    def forward(self, x):
        bs, nc, nx, ny, nz  = x.shape
        x1 = self.encoder1(x)
        x2 = self.encoder2(x1)
        x3 = self.encoder3(x2)
        x4 = self.encoder4(x3)
        x5 = self.encoder5(x4)
        x4_ = self.decoder1_conv(torch.cat((x4, self.decoder1_trans(x5)), dim = 1))
        x3_ = self.decoder2_conv(torch.cat((x3, self.decoder2_trans(x4_)), dim = 1))
        x2_ = self.decoder3_conv(torch.cat((x2, self.decoder3_trans(x3_)), dim = 1))
        x1_ = self.decoder4_conv(torch.cat((x1, self.decoder4_trans(x2_)), dim = 1))

        if not self.is_binary:
            x1_ = F.softmax(self.final_layer1(x1_).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x2_ = F.softmax(self.upsample2(self.final_layer2(x2_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x3_ = F.softmax(self.upsample3(self.final_layer3(x3_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x4_ = F.softmax(self.upsample4(self.final_layer4(x4_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
        else:
            x1_ = torch.sigmoid(self.final_layer1(x1_).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x2_ = torch.sigmoid(self.upsample2(self.final_layer2(x2_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x3_ = torch.sigmoid(self.upsample3(self.final_layer3(x3_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x4_ = torch.sigmoid(self.upsample4(self.final_layer4(x4_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
        
        result = torch.stack([x1_, x2_, x3_, x4_]).permute(1,0,2,3,4,5)
        
        return result.contiguous()
                
    def get_networks_depth(self):
        min_patch_axis = np.min(self.patch_size)
        return np.clip(np.log2(min_patch_axis), 0, 6).astype(int)
    

class Deep_Supervision_UNet3d(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, patch_size = (128, 128, 128), is_binary = False):
        super().__init__()
        self.patch_size = np.array(patch_size)
        self.networks_depth = self.get_networks_depth()
        self.encoder_blocks = []
        self.decoder_blocks = []
        self.upsample_operations = []
        self.final_output_channels = out_channels
        self.is_binary = is_binary
        
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        self.encoder1 = nn.Sequential(
            ConvDropoutNormNonlin(in_channels, 32, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(32, 32, conv_kwargs = conv_kwargs1, conv_op = nn.Conv3d)
        )
        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 2}
        conv_kwargs2 = {'kernel_size' : 3, 'padding' : 1, 'stride' : 1}
        self.encoder2 = nn.Sequential(
            ConvDropoutNormNonlin(32, 64, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(64, 64, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder3 = nn.Sequential(
            ConvDropoutNormNonlin(64, 128, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(128, 128, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder4 = nn.Sequential(
            ConvDropoutNormNonlin(128, 256, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(256, 256, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.encoder5 = nn.Sequential(
            ConvDropoutNormNonlin(256, 320, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(320, 320, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )

        conv_kwargs1 = {'kernel_size' : 3, 'padding' : 1,  'stride' : 1}
        conv_kwargs2 = {'kernel_size' : 3, 'stride' : 1, 'padding':1}
        transp_kwargs = {'kernel_size' : 2, 'stride' : 1, 'stride' : 2}


        self.decoder1_trans = ConvDropoutNormNonlin(320, 256, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder1_conv = nn.Sequential(
            ConvDropoutNormNonlin(256*2, 256, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(256, 256, conv_kwargs = conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder2_trans = ConvDropoutNormNonlin(256, 128, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder2_conv = nn.Sequential(
            ConvDropoutNormNonlin(128*2, 128 , conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(128, 128, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder3_trans = ConvDropoutNormNonlin(128, 64, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder3_conv = nn.Sequential(
            ConvDropoutNormNonlin(64*2, 64, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(64, 64, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )
        self.decoder4_trans = ConvDropoutNormNonlin(64, 32, conv_kwargs= transp_kwargs, conv_op = nn.ConvTranspose3d)
        self.decoder4_conv = nn.Sequential(
            ConvDropoutNormNonlin(32*2, 32, conv_kwargs= conv_kwargs1, conv_op = nn.Conv3d),
            ConvDropoutNormNonlin(32, 32, conv_kwargs= conv_kwargs2, conv_op = nn.Conv3d)
        )

        self.final_layer1 = nn.Sequential(
            nn.Conv3d(32, out_channels, kernel_size = 1, stride = 1, padding = 0)
            )
        self.final_layer2 = nn.Sequential(
            nn.Conv3d(64, out_channels, kernel_size = 1, stride = 1, padding = 0)
            )
        self.final_layer3 = nn.Sequential(
            nn.Conv3d(128, out_channels, kernel_size = 1, stride = 1, padding = 0)
            )
        self.final_layer4 = nn.Sequential(
            nn.Conv3d(256, out_channels, kernel_size = 1, stride = 1, padding = 0)
            )

        self.upsample4 = nn.Upsample(scale_factor=8, mode = 'trilinear')
        self.upsample3 = nn.Upsample(scale_factor=4, mode = 'trilinear')
        self.upsample2 = nn.Upsample(scale_factor=2, mode = 'trilinear')
        
    
    def forward(self, x):
        bs, nc, nx, ny, nz  = x.shape
        x1 = self.encoder1(x)
        x2 = self.encoder2(x1)
        x3 = self.encoder3(x2)
        x4 = self.encoder4(x3)
        x5 = self.encoder5(x4)

        x4_ = self.decoder1_conv(torch.cat((x4, self.decoder1_trans(x5)), dim = 1))
        x3_ = self.decoder2_conv(torch.cat((x3, self.decoder2_trans(x4_)), dim = 1))
        x2_ = self.decoder3_conv(torch.cat((x2, self.decoder3_trans(x3_)), dim = 1))
        x1_ = self.decoder4_conv(torch.cat((x1, self.decoder4_trans(x2_)), dim = 1))

        if not self.is_binary:
            x1_ = F.softmax(self.final_layer1(x1_).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x2_ = F.softmax(self.upsample2(self.final_layer2(x2_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x3_ = F.softmax(self.upsample3(self.final_layer3(x3_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
            x4_ = F.softmax(self.upsample4(self.final_layer4(x4_)).view(bs, self.final_output_channels, -1), dim = 1).view(bs, self.final_output_channels, nx, ny, nz)
        

        else:
            x1_ = torch.sigmoid(self.final_layer1(x1_).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x2_ = torch.sigmoid(self.upsample2(self.final_layer2(x2_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x3_ = torch.sigmoid(self.upsample3(self.final_layer3(x3_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
            x4_ = torch.sigmoid(self.upsample4(self.final_layer4(x4_)).view(bs, self.final_output_channels, -1)).view(bs, self.final_output_channels, nx, ny, nz)
        
        result = torch.stack([x1_, x2_, x3_, x4_]).permute(1,0,2,3,4,5)
        
        return result.contiguous()
                
    def get_networks_depth(self):
        min_patch_axis = np.min(self.patch_size)
        return np.clip(np.log2(min_patch_axis), 0, 6).astype(int)
    


class Interactive_Model(nn.Module):
    def __init__(self, n_modalities = 1, point_channel = 2, patch_size = (320,256,256), final_output_channels = 1):
        super().__init__()
        kwargs = {'sample_input_D':patch_size[0], 'sample_input_H':patch_size[1], 'sample_input_W':patch_size[2],'num_seg_classes':1}
        self.model = resnet10(**kwargs)
        input = n_modalities + final_output_channels + point_channel

        self.point_layers = nn.Sequential(
            nn.Conv3d(1, point_channel, 1, 1, 0),
            nn.InstanceNorm3d(point_channel),
            nn.LeakyReLU(),
            nn.Conv3d(point_channel, point_channel, 3, 1, 1),
            nn.InstanceNorm3d(point_channel),
            nn.LeakyReLU(),
            
        )
        self.initial_layer = nn.Sequential(
            nn.Conv3d(input, 3, 3, 1, 1),
            nn.InstanceNorm3d(32),
            nn.LeakyReLU(),
            nn.Conv3d(3, 1, 3, 1, 1),
            nn.InstanceNorm3d(final_output_channels),
            nn.LeakyReLU(),
            )
        self.final_upsample = nn.Upsample(scale_factor=4, mode = 'trilinear')
        
        
    
    def forward(self, x, positive_points, negative_points, prev_seg = None):
        bs, nc, nx, ny, nz  = x.shape
        if prev_seg == None:
            prev_seg = torch.zeros((bs, 1, nx, ny, nz)).to(x.device.index)
        point = self.point_layers((positive_points - negative_points).unsqueeze(1))
        
        x = torch.cat((x, point), dim = 1)
        x = torch.cat((x, prev_seg), dim = 1)
        x = self.initial_layer(x)
        x = self.model(x)
        x = self.final_upsample(x)

        return torch.sigmoid(x)




class Daegu_Model10(nn.Module):
    def __init__(self, patch_size, resnet_output = 4):
        super().__init__()
        resnet_output = resnet_output
        out_tconv1 = resnet_output//2
        out_tconv2 = out_tconv1//2
        kwargs = {'sample_input_D':patch_size[0], 'sample_input_H':patch_size[1], 'sample_input_W':patch_size[2],'num_seg_classes':resnet_output}
        self.model = resnet10(**kwargs)

        self.upsample1 = nn.Sequential(nn.ConvTranspose3d(in_channels = resnet_output, out_channels= out_tconv1, kernel_size =3 , stride = 2, padding = 1, output_padding =1),
                                       nn.InstanceNorm3d(out_tconv1),
                                       nn.LeakyReLU())
        
        self.upsample2 = nn.Sequential(nn.ConvTranspose3d(in_channels = out_tconv1, out_channels= out_tconv2, kernel_size =3 , stride = 2, padding = 1, output_padding =1),
                                       nn.InstanceNorm3d(out_tconv2),
                                       nn.LeakyReLU())
        
        self.final_layer = nn.Sequential(nn.Conv3d(in_channels = out_tconv2, out_channels = 2, kernel_size = 3, stride = 1, padding = 1),
                                   nn.InstanceNorm3d(2),
                                   nn.LeakyReLU())

    def forward(self, X):
        X = self.model(X)
        X = self.upsample1(X)
        X = self.upsample2(X)
        X = self.final_layer(X)
        return torch.softmax(X, dim = -1)

class Daegu_Model34(nn.Module):
    def __init__(self, patch_size, input_channel = 1, output_channel= 2):
        super().__init__()
        out_channel1 = 1
        out_channel2 = out_channel1 * 4

        self.a_conv1 = nn.Sequential(nn.Conv3d(in_channels = input_channel, out_channels = out_channel1, kernel_size = 3, stride = 1, padding = 6, dilation = 6),
                                   nn.InstanceNorm3d(out_channel1),
                                   nn.LeakyReLU())
        self.a_conv2 = nn.Sequential(nn.Conv3d(in_channels = input_channel, out_channels = out_channel1, kernel_size = 3, stride = 1, padding = 12, dilation = 12),
                                   nn.InstanceNorm3d(out_channel1),
                                   nn.LeakyReLU())
        self.a_conv3 = nn.Sequential(nn.Conv3d(in_channels = input_channel, out_channels = out_channel1, kernel_size = 3, stride = 1, padding = 18, dilation = 18),
                                   nn.InstanceNorm3d(out_channel1),
                                   nn.LeakyReLU())
        self.a_conv4 = nn.Sequential(nn.Conv3d(in_channels = input_channel, out_channels = out_channel1, kernel_size = 1, stride = 1, padding = 0, dilation = 1),
                                   nn.InstanceNorm3d(out_channel1),
                                   nn.LeakyReLU())

        self.a_conv2_1 = nn.Sequential(nn.Conv3d(in_channels = out_channel2, out_channels = out_channel2, kernel_size = 1, stride = 1, padding = 0, dilation = 1),
                                   nn.InstanceNorm3d(out_channel2),
                                   nn.LeakyReLU())

        self.last_layer = nn.Sequential(nn.Conv3d(in_channels = out_channel2, out_channels = output_channel, kernel_size = 1, stride = 1, padding = 0, dilation = 1),
                                   nn.InstanceNorm3d(output_channel),
                                   nn.LeakyReLU())


    def forward(self, X):
        X1 = self.a_conv1(X)
        X2 = self.a_conv2(X)
        X3 = self.a_conv3(X)
        X4 = self.a_conv4(X)

        X = torch.cat((X1,X2,X3,X4), dim = 1)
        X = self.a_conv2_1(X)
        X = self.last_layer(X)
        return torch.softmax(X, dim = -1)




def conv3x3x3(in_planes, out_planes, stride=1, dilation=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        dilation=dilation,
        stride=stride,
        padding=dilation,
        bias=False)


def downsample_basic_block(x, planes, stride, no_cuda=False):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3),
        out.size(4)).zero_()
    if not no_cuda:
        if isinstance(out.data, torch.cuda.FloatTensor):
            zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3x3(inplanes, planes, stride=stride, dilation=dilation)
        self.bn1 = nn.BatchNorm3d(planes)
        self.relu = nn.LeakyReLU(inplace = True)
        self.conv2 = conv3x3x3(planes, planes, dilation=dilation)
        self.bn2 = nn.BatchNorm3d(planes)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv3d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm3d(planes)
        self.conv2 = nn.Conv3d(
            planes, planes, kernel_size=3, stride=stride, dilation=dilation, padding=dilation, bias=False)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes * 4)
        self.relu = nn.LeakyReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self,
                 block,
                 layers,
                 sample_input_D,
                 sample_input_H,
                 sample_input_W,
                 num_seg_classes,
                 shortcut_type='B',
                 no_cuda = False):
        self.inplanes = 64
        self.no_cuda = no_cuda
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv3d(
            1,
            64,
            kernel_size=7,
            stride=(2, 2, 2),
            padding=(3, 3, 3),
            bias=False)
            
        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.LeakyReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], shortcut_type)
        self.layer2 = self._make_layer(
            block, 128, layers[1], shortcut_type, stride=2)
        self.layer3 = self._make_layer(
            block, 256, layers[2], shortcut_type, stride=1, dilation=2)
        self.layer4 = self._make_layer(
            block, 512, layers[3], shortcut_type, stride=1, dilation=4)

        self.conv_seg = nn.Sequential(
                                        nn.ConvTranspose3d(
                                        512 * block.expansion,
                                        32,
                                        2,
                                        stride=2
                                        ),
                                        nn.BatchNorm3d(32),
                                        nn.LeakyReLU(inplace=True),
                                        nn.Conv3d(
                                        32,
                                        32,
                                        kernel_size=3,
                                        stride=(1, 1, 1),
                                        padding=(1, 1, 1),
                                        bias=False), 
                                        nn.BatchNorm3d(32),
                                        nn.LeakyReLU(inplace=True),
                                        nn.Conv3d(
                                        32,
                                        num_seg_classes,
                                        kernel_size=1,
                                        stride=(1, 1, 1),
                                        bias=False) 
                                        )

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                m.weight = nn.init.kaiming_normal(m.weight, mode='fan_out')

    def _make_layer(self, block, planes, blocks, shortcut_type, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if shortcut_type == 'A':
                downsample = partial(
                    downsample_basic_block,
                    planes=planes * block.expansion,
                    stride=stride,
                    no_cuda=self.no_cuda)
            else:
                downsample = nn.Sequential(
                    nn.Conv3d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False), nn.BatchNorm3d(planes * block.expansion))

        layers = []
        layers.append(block(self.inplanes, planes, stride=stride, dilation=dilation, downsample=downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.conv_seg(x)

        return x


def resnet10(**kwargs):
    """Constructs a ResNet-10 model.
    """
    model = ResNet(BasicBlock, [1, 1, 1, 1], **kwargs)
    return model


def resnet18(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    return model


def resnet34(**kwargs):
    """Constructs a ResNet-34 model.
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    return model


def resnet50(**kwargs):
    """Constructs a ResNet-50 model.
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    return model


def resnet152(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    return model


def resnet200(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 24, 36, 3], **kwargs)
    return model


from ingradient_library.dataloads import *
from ingradient_library.inference import Inference
from ingradient_library.preprocessing import *
from ingradient_library.dataloads import DataLoader3D
from ingradient_library.dataloads import CustomDataset
from ingradient_library.preprocessing import Resampling
from ingradient_library.preprocessing import Normalizer
from ingradient_library.preprocessing import Get_target_spacing
from ingradient_library.transform import Transform
from ingradient_library.get_nnunet_setting import get_transform_params
from ingradient_library.deep_supervision_loss import *
from ingradient_library.trainer import Trainer
from ingradient_library.model import *
from ingradient_library.model import *
from ingradient_library.get_information import get_imbalance_weight
from ingradient_library.optimizer import SAMSGD
from torch.utils.data.dataset import random_split
from ingradient_library.get_information import get_imbalance_weight
from ingradient_library.active_contour_loss import *
import torch.nn.functional as F
import torch.optim as optim
import torch.nn as nn


torch.manual_seed(1)

def nnunet_3D_run(LOAD_PATH, SAVE_MODEL_PATH, GPU_DEVICE_INDEX, patch_size, batch_size, n_classes, n_modalities):
    transform = Transform(*get_transform_params(GPU_DEVICE_INDEX))
    dataset = CustomDataset(LOAD_PATH, normalizer = Normalizer())
    tr_dataset, val_dataset = random_split(dataset, [len(dataset) - len(dataset)//9, len(dataset)//9])
    tr_dataloader = DataLoader3D(tr_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size, transform = transform)
    val_dataloader = DataLoader3D(val_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size)
    weight = get_imbalance_weight(LOAD_PATH, n_classes)

    model = Deep_Supervision_UNet3d(patch_size = patch_size, n_modalities = n_modalities, final_output_channels = n_classes, is_binary = False).to(GPU_DEVICE_INDEX)
    loss2 = Multi_Dice_Loss(n_classes)
    loss1 = nn.CrossEntropyLoss(reduction = 'none', weight = torch.tensor(weight).float().to(GPU_DEVICE_INDEX))
    optimizer = optim.SGD(model.parameters(), lr = 1e-2, nesterov = True, momentum = 0.99)
    scheduler = optim.lr_scheduler.LambdaLR(optimizer=optimizer,
                                lr_lambda=lambda epoch: 0.95 ** epoch)
    trainer = Trainer(tr_dataloader = tr_dataloader, val_dataloader= val_dataloader, model =  model, optimizer = optimizer, scheduler = scheduler, losses = [loss1, loss2], save_path = SAVE_MODEL_PATH, is_deep_supervision = True)
    trainer.run()



def nnunet_3D_run_narrow(LOAD_PATH, SAVE_MODEL_PATH, GPU_DEVICE_INDEX, patch_size, batch_size, n_classes, n_modalities):
    transform = Transform(*get_transform_params(GPU_DEVICE_INDEX))
    dataset = CustomDataset(LOAD_PATH, normalizer = Normalizer())
    print(dataset[0][0].shape)
    tr_dataset, val_dataset = random_split(dataset, [len(dataset) - len(dataset)//9, len(dataset)//9])
    tr_dataloader = DataLoader3D(tr_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size, transform = transform)
    val_dataloader = DataLoader3D(val_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size)
    weight = get_imbalance_weight(LOAD_PATH, n_classes)

    model = Deep_Supervision_UNet3d_Narrow(patch_size = patch_size, n_modalities = n_modalities, final_output_channels = n_classes, is_binary = False).to(GPU_DEVICE_INDEX)
    loss2 = Multi_Dice_Loss(n_classes)
    loss1 = nn.CrossEntropyLoss(reduction = 'none', weight = torch.tensor(weight).float().to(GPU_DEVICE_INDEX))
    optimizer = optim.Adam(model.parameters(), lr = 1e-2)
    scheduler = optim.lr_scheduler.LambdaLR(optimizer=optimizer,
                                lr_lambda=lambda epoch: 0.95 ** epoch)
    trainer = Trainer(tr_dataloader = tr_dataloader, val_dataloader= val_dataloader, model =  model, optimizer = optimizer, scheduler = scheduler, losses = [loss1, loss2], save_path = SAVE_MODEL_PATH, is_deep_supervision = True)
    trainer.run()



def narrow_inference_run(LOAD_PATH, MODEL_LOAD_PATH,  SAVE_RESULT_PATH, GPU_DEVICE_INDEX, patch_size, n_classes, n_modalities):
    dataset = CustomDataset(LOAD_PATH, normalizer = Normalizer())
    _, val_dataset = random_split(dataset, [len(dataset) - len(dataset)//9, len(dataset)//9])
    ifr = Inference(dataset = val_dataset, n_classes = n_classes, patch_size = patch_size, device = GPU_DEVICE_INDEX, save_path=SAVE_RESULT_PATH)
    model = Deep_Supervision_UNet3d_Narrow(patch_size = patch_size, n_modalities = n_modalities, final_output_channels = n_classes, is_binary = False).to(GPU_DEVICE_INDEX)
    file = torch.load(MODEL_LOAD_PATH)
    model.load_state_dict(file)
    ifr.run(model, mode='dice')
    print(ifr.get_result())


def nnunet_lung_run(LOAD_PATH, SAVE_MODEL_PATH, GPU_DEVICE_INDEX, patch_size = (208, 208, 112), batch_size = 2, n_classes = 2, n_modalities = 1):
    transform = Transform(*get_transform_params(GPU_DEVICE_INDEX))
    resample = Resampling(target_spacing = (0.79, 0.79, 1.24), anisotropy_axis_index=np.array([]))
    dataset = CustomDataset(LOAD_PATH, normalizer = Fixed_Normalizer(), resample = resample)
    print(dataset[0][0].shape, patch_size)
    tr_dataset, val_dataset = random_split(dataset, [len(dataset) - len(dataset)//9, len(dataset)//9])
    tr_dataloader = DataLoader3D(tr_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size, transform = transform)
    val_dataloader = DataLoader3D(val_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size)
    model = Deep_Supervision_UNet3d_Lung(patch_size = patch_size, n_modalities = n_modalities, final_output_channels = n_classes, is_binary = False).to(GPU_DEVICE_INDEX)
    optimizer = SAMSGD(model.parameters(), lr = 1e-9, nesterov = True, momentum = 0.99, weight_decay= 0.05)
    
    writer = SummaryWriter()
    current_iter = 0
    val_iter = 0
    criterion1 = FastACELoss3DV2(classes = n_classes).to(GPU_DEVICE_INDEX)
    criterion2 = nn.CrossEntropyLoss(reduction = 'sum', weight = torch.tensor([1,39500]).float().to(GPU_DEVICE_INDEX))
    step = 0
    v_step = 0
    for e in range(1000):
        tr_dataloader.new_epoch()
        n_iter = 0
        model.train()
        epoch_tr_loss = 0
        epoch_tr_div = 0
        epoch_val_loss = 0
        epoch_val_div = 0
        while not tr_dataloader.is_end():
            current_iter += 1
            n_iter += 1
            images, seg = tr_dataloader.generate_train_batch()
            if isinstance(optimizer, SAMSGD):
                def closure():
                    optimizer.zero_grad()
                    output = model(images)
                    output = output.permute(0,2,3,4,1)
                    output = torch.softmax(output, dim = -1)
                    output = output.permute(0,-1,1,2,3)
                    loss1 = criterion1(output, F.one_hot(seg,2).permute(0,-1,1,2,3))
                    loss2 = criterion2(output, seg) / 1e+3
                    writer.add_scalar('Lung/Train/AC Loss', loss1.item(), step)
                    writer.add_scalar('Lung/Train/CE Loss', loss2.item() * 1e+3, step)
                    loss = loss1 + loss2
                    loss.backward()
                    return loss
                optimizer.step(closure)
                epoch_tr_loss = 1
                epoch_tr_div = 1
            step+=1

       
        val_dataloader.new_epoch()
        val_loss = 0
        n_iter = 0
        model.eval()
        while not val_dataloader.is_end():
            val_iter +=1
            n_iter +=1
            with torch.no_grad():
                images, seg = val_dataloader.generate_train_batch()
                output = model(images)
                output = output.permute(0,2,3,4,1)
                output = torch.softmax(output, dim = -1)
                output = output.permute(0,-1,1,2,3)
                loss1 = criterion1(output, F.one_hot(seg,2).permute(0,-1,1,2,3))
                loss2 = criterion2(output, seg) / 1e+3
                writer.add_scalar('Lung/Valid/AC Loss', loss1.item(), step)
                writer.add_scalar('Lung/Valid/CE Loss', loss2.item(), step)
                loss = loss1 + loss2
                epoch_val_loss += loss.item()
                epoch_val_div += 1
            v_step += 1


        writer.add_scalar('epoch\Loss_train', epoch_tr_loss/epoch_tr_div , e)
        writer.add_scalar('epoch\Loss_val', epoch_val_loss/epoch_val_div , e) 

        file_name = 'epoch'+ str(e) + '_model_state_dict.pkl'
        torch.save(model.state_dict(), os.path.join(SAVE_MODEL_PATH, file_name))


def nnunet_brain_run(LOAD_PATH, SAVE_MODEL_PATH, GPU_DEVICE_INDEX, patch_size = (144, 160, 144), batch_size = 2, n_classes = 5, n_modalities = 4):
    transform = Transform(*get_transform_params(GPU_DEVICE_INDEX))
    dataset = CustomDataset(LOAD_PATH, normalizer = Normalizer())
    print(dataset[0][0].shape, patch_size)
    tr_dataset, val_dataset = random_split(dataset, [len(dataset) - len(dataset)//9, len(dataset)//9])
    tr_dataloader = DataLoader3D(tr_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size, transform = transform)
    val_dataloader = DataLoader3D(val_dataset, patch_size = patch_size, device = GPU_DEVICE_INDEX, batch_size = batch_size)
    weight = get_imbalance_weight(LOAD_PATH, n_classes)

    model = Deep_Supervision_UNet3d(patch_size = patch_size, n_modalities = n_modalities, final_output_channels = n_classes, is_binary = False).to(GPU_DEVICE_INDEX)
    loss2 = Multi_Dice_Loss(n_classes)
    loss1 = nn.CrossEntropyLoss(reduction = 'none', weight = torch.tensor(weight).float().to(GPU_DEVICE_INDEX))
    optimizer = optim.Adam(model.parameters(), lr = 1e-2)
    scheduler = optim.lr_scheduler.LambdaLR(optimizer=optimizer,
                                lr_lambda=lambda epoch: 0.95 ** epoch)
    trainer = Trainer(tr_dataloader = tr_dataloader, val_dataloader= val_dataloader, model =  model, optimizer = optimizer, scheduler = scheduler, losses = [loss1, loss2], save_path = SAVE_MODEL_PATH, is_deep_supervision = True)
    trainer.run()



from torch.utils.data import Dataset, DataLoader
from ingradient_library.transform import Transform
from ingradient_library.preprocessing import *
from ingradient_library.get_nnunet_setting import *
from ingradient_library.patch_transform import *
from ingradient_library.unet import *
import torch.nn as nn
import torch.nn.functional as F
import torch
import numpy as np
import os
import h5py
import copy
import numpy as np
import cc3d


class MAIC_Dataset(Dataset):
    def __init__(self, path = None, normalizer = True, train = True, transform = Transform(*get_maic_params(None))):
        if path == None:
            path = '../mnt/dataset/'
        
        self.path = path
        self.file_list = []
        for f in os.listdir(path):
            if not 'py' in f:
                self.file_list.append(f)
        
        self.file_list = sorted(self.file_list)
        self.normalizer = normalizer
        self.train = train
        self.sampler = MAIC_Sampling(transform = transform, train = train)
        self.norm1 = Fixed_Normalizer(mean = 20.78, std = 180.50, min = -986, max = 271,  device = None)
        #self.norm1 = Fixed_Normalizer(mean = -410, std = 400,  device = None)
        
    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        current_file = os.path.join(self.path, self.file_list[idx])
        hdf_file = h5py.File(current_file , 'r')
        CT = np.array(hdf_file['CT'])
        PET = np.array(hdf_file['PET'])
        spacing = np.array(hdf_file['Size'])
        
        if self.train:
            seg = np.array(hdf_file['Aorta'])
            images = np.expand_dims(CT ,axis = 0)
        
        else:
            images = np.stack((CT, PET))
            seg = None
        
        hdf_file.close()
        images, seg = self.sampler(images, seg)
        if self.normalizer:
            images[0] = self.norm1(images[0])
        
        if not self.train:
            PET = images[1]
            images = np.expand_dims(images[0], axis = 0)
            return images, seg, spacing, PET
        else:
            return images, seg, spacing, 1
        
class CT_Lung_Cropping(object):
    def __init__(self, clip = False, threshold_clip = [-900, -200 ], threshold_cut = [-1000, -350]):
        self.clip = clip
        self.threshold_clip = threshold_clip
        self.threshold_cut = threshold_cut
    
    def __call__(self, image):
        if self.clip:
            clip_img = np.clip(image, self.threshold_clip[0], self.threshold_clip[1])
        else:
            clip_img = image
        mask =  (clip_img > self.threshold_cut[0]) * (clip_img < self.threshold_cut[1])
        if not isinstance(mask, np.ndarray):
            mask = mask.numpy()
        conected_component_mask = cc3d.connected_components(mask, connectivity = 26)
        count = np.unique(conected_component_mask, return_counts = True)[1]
        lung_index = np.argsort(count)[::-1]
        if len(lung_index) > 2:
            lung_index = lung_index[2]
        
        else:
            lung_index = lung_index[-1]
        lung_max = np.max(np.where(conected_component_mask == lung_index), axis = 1)
        lung_min = np.min(np.where(conected_component_mask == lung_index), axis = 1)
        
        return lung_min, lung_max


class MAIC_Sampling(object):
    def __init__(self,  train = True, transform = None):
        self.transform = transform
        self.train = train
        self.lung_cut = CT_Lung_Cropping()
        
    def __call__(self, images, seg = None, is_CT = True):
        if not isinstance(images, np.ndarray):
            images = np.array(images)
            if self.train:
                seg = np.array(seg)
        
        
        temp = copy.deepcopy(images)
        if is_CT:
            temp[np.where(temp < - 700)] = 0
        
        non_zero_index = np.where(temp.astype(int) != 0)
        min_val = np.min(non_zero_index, axis = 1)
        max_val = np.max(non_zero_index, axis = 1)
        if self.train:
            random_move = np.random.randint([-5,-5,-5], [5, 5, 5])
        else:
            random_move = np.array([0,0,0])
        random_move = np.array([0,0,0])
        images = images[:, min_val[-3]:max_val[-3]+1, min_val[-2]:max_val[-2]+1, min_val[-1]:max_val[-1]+1]
        z_start = int(images.shape[-1] * 0.25)
        z_term = 128
        y_start = images.shape[-2]//2 - 40
        y_end= images.shape[-2]//2 + 40
        x_term = 128
        images = images[:, 5 + random_move[0]:x_term+random_move[0]+5, y_start+random_move[1]:y_end+random_move[1],
                       -(z_start + z_term) + random_move[2]:-z_start + random_move[2]]
        
        lung_index = self.lung_cut(images[0])
        lung_center = (lung_index[0] + lung_index[1])//2
        lung_center[2] = min(lung_center[2], 128 - 48)
        images = images[:, lung_center[0]-24:lung_center[0]+24, lung_center[1] - 24: lung_center[1] + 24, lung_center[2]-48:lung_center[2]+48]
        if self.train:
            seg = seg[min_val[-3]:max_val[-3]+1, min_val[-2]:max_val[-2]+1, min_val[-1]:max_val[-1]+1]
            seg = seg[5 + random_move[0]:x_term+random_move[0]+5, y_start+random_move[1]:y_end+random_move[1],
                       -(z_start + z_term) + random_move[2]:-z_start + random_move[2]]
            seg = seg[lung_center[0]-24:lung_center[0]+24, lung_center[1] - 24: lung_center[1] + 24, lung_center[2]-48:lung_center[2]+48]
            images = torch.tensor(images).unsqueeze(0).double()
            seg = torch.tensor(seg).unsqueeze(0).long()
            if self.transform != None:
                images, seg = self.transform(images, seg, None)
            images = images.squeeze(0).numpy()
            seg = seg.squeeze(0).numpy()
        else: 
            seg = None

        return images, seg


class MU_Dataset(Dataset):
    def __init__(self, path = None, normalizer = True, train = True, transform = None):
        if path == None:
            path = '../mnt/dataset/'
        
        self.path = path
        self.file_list = []
        for f in os.listdir(path):
            if not 'py' in f:
                self.file_list.append(f)
        
        self.file_list = sorted(self.file_list)
        self.normalizer = normalizer
        self.train = train
        self.sampler = MAIC_Sampling(transform = transform, train = train)
        self.norm1 = Fixed_Normalizer(mean = 20.78, std = 180.50, min = -986, max = 271,  device = None)
        self.norm2 = Fixed_Normalizer(mean = 0.004, std = 0.007,  device = None)
        
    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        current_file = os.path.join(self.path, self.file_list[idx])
        hdf_file = h5py.File(current_file , 'r')
        CT = np.array(hdf_file['CT'])
        PET = np.array(hdf_file['PET'])
        spacing = np.array(hdf_file['Size'])
        seg = np.array(hdf_file['Aorta'])
        images = np.stack((CT, PET))
        hdf_file.close()
        images, seg = self.sampler(images, seg)
        
        if self.normalizer:
            """
            images[0] = self.norm1(images[0])
            
            images[1] = (images[1] - 0.7706) / 0.6547
            """
            pet = images[1].copy()
            #images[0] = (images[0] - 1275) / 995
            #images[0][np.where(images[0] < 0)] = -200
            #images[0] = np.clip(images[0], None, 200)
            #images[1] = (images[1] - 0.14) / 0.75
            images[0] = (images[0] + 253.7455) / 417.6291
        return np.expand_dims(images[0], axis = 0), seg, spacing, np.expand_dims(images[1], axis = 0), pet


from setuptools import setup, find_packages
 
setup(
    name                = 'ingradient_lib_temp',
    version             = '0.6.247',
    description         = 'Medical Deep Learning Framework.',
    author              = 'seungyeob.seon',
    author_email        = 'liamseon@gmail.com',
    url                 = 'https://github.com/InGradient/InGradient_AI_Library',
    license             = 'MIT',
    install_requires    =  ['SimpleITK', 'revlib', 'connected-components-3d', 'nibabel'],
    py_modules          = ['data_augmentation', 'dataloads', 'deep_supervision_loss', 'get_imbalance_weight'
                          ,'inference', 'patch_transform', 'preprocessing', 'model', 'visualization', 'sampling'
                          ,'preprocessing', 'trainer', 'transform', 'data_organizer', 'get_nnunet_setting', 'optimizer', 'active_contour_loss'
                          ,'medical_decathlon_organizer', 'nnunet_3D_run', 'maic', 'unet', 'lr_scheduler', 'loss'],
    keywords            = ['pypi deploy'],
    packages            = ['ingradient_library'],
    zip_safe            = False
)



import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint_sequential
from ingradient_library.block import *


class CheckPointNet(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 32, segment = 8):
        super().__init__()
        self.model = nn.Sequential(nn.Conv3d(in_channels, n_channels, kernel_size = 3, stride = 1, padding = 1, dilation = 1),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 3, stride = 1, padding = 24, dilation = 24),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 3, stride = 1, padding = 18, dilation = 18),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 3, stride = 1, padding = 12, dilation = 12),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 3, stride = 1, padding = 6, dilation = 6),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 3, stride = 1, padding = 1, dilation = 1),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, n_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1),
                        nn.Dropout3d(),
                        nn.InstanceNorm3d(n_channels),
                        nn.LeakyReLU(),
                        nn.Conv3d(n_channels, out_channels, kernel_size = 1, stride = 1, padding = 0, dilation = 1)
                        )
        self.segment = segment

    def forward(self, X):
        if self.segment != 0:
            X = checkpoint_sequential(self.model, self.segment, X)
        
        else:
            X = self.model(X)

        return X

class UNet3d_Plus(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 32, channel_bound = 320, vectorization = True):
        super().__init__()
        self.expand_channels = encoder_block(in_channels, n_channels, stride = 1)
        self.conv_downsample1 = encoder_block(n_channels, n_channels * 2)
        self.conv_downsample2 = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3 = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4 = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1 = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1 = encoder_block(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2 = encoder_block(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3 = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4 = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv3d(n_channels, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv3d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv3d(n_channels * 4, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv3d(n_channels * 8, out_channels, kernel_size = 1)

        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)
        self.vectorization = vectorization

    
    def forward(self, x):
        x = self.expand_channels(x)
        x_downsample1 = self.conv_downsample1(x)
        x_downsample2 = self.conv_downsample2(x_downsample1)
        x_downsample3 = self.conv_downsample3(x_downsample2)
        x_downsample4 = self.conv_downsample4(x_downsample3)

        x_downsample4_up = self.conv_upsample1(x_downsample4)
        l4_output = self.decoder4(torch.cat((x_downsample4_up, x_downsample3), dim = 1))
        x_downsample3_up = self.conv_upsample2(l4_output)
        l3_output = self.decoder3(torch.cat((x_downsample3_up, x_downsample2), dim = 1))
        x_downsample2_up = self.conv_upsample3(l3_output)
        l2_output = self.decoder2(torch.cat((x_downsample2_up, x_downsample1), dim = 1))
        x_downsample1_up = self.conv_upsample4(l2_output)
        l1_output = self.decoder1(torch.cat((x_downsample1_up, x), dim = 1))

        l1_output = self.classifier1(l1_output)
        l2_output = self.upsample2(self.classifier2(l2_output))
        l3_output = self.upsample3(self.classifier3(l3_output))
        l4_output = self.upsample4(self.classifier4(l4_output))

        if self.vectorization:
            l1_output = torch.softmax(l1_output , dim = 1)
            l2_output = torch.softmax(l2_output , dim = 1)
            l3_output = torch.softmax(l3_output , dim = 1)
            l4_output = torch.softmax(l4_output , dim = 1)
            result = torch.stack([l1_output, l2_output, l3_output, l4_output]).permute(1,0,2,3,4,5)
            
            return result.contiguous()

        
        return [l1_output, l2_output, l3_output, l4_output]



class UNet3d_MultiModal(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 32, channel_bound = 320):
        super().__init__()
        channel_bound = min(n_channels * 16, channel_bound)
        self.expand_channels_CT = encoder_block(in_channels, n_channels, stride = 1)
        self.conv_downsample1_CT = encoder_block(n_channels, n_channels * 2)
        self.conv_downsample2_CT = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3_CT = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4_CT = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1_CT = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2_CT= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3_CT= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4_CT= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1_CT = encoder_block(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2_CT = encoder_block(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3_CT = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4_CT = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)

        self.expand_channels_PET = encoder_block(in_channels, n_channels, stride = 1)
        self.conv_downsample1_PET = encoder_block(n_channels, n_channels * 2)
        self.conv_downsample2_PET = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3_PET = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4_PET = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1_PET = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2_PET= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3_PET= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4_PET= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1_PET = encoder_block(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2_PET = encoder_block(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3_PET = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4_PET = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv3d(n_channels, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv3d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv3d(n_channels * 4, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv3d(n_channels * 8, out_channels, kernel_size = 1)

        self.classifier_PET = nn.Conv3d(n_channels, 1, kernel_size = 1)

        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)


    def forward(self, CT, PET):
        PET = self.expand_channels_PET(PET)
        PET_downsample1 = self.conv_downsample1_PET(PET)
        PET_downsample2 = self.conv_downsample2_PET(PET_downsample1)
        PET_downsample3 = self.conv_downsample3_PET(PET_downsample2)
        PET_downsample4 = self.conv_downsample4_PET(PET_downsample3)

        PET_downsample4_up = self.conv_upsample1_PET(PET_downsample4)
        l4_output_PET = self.decoder4_PET(torch.cat((PET_downsample4_up, PET_downsample3), dim = 1))
        PET_downsample3_up = self.conv_upsample2_PET(l4_output_PET)
        l3_output_PET = self.decoder3_PET(torch.cat((PET_downsample3_up, PET_downsample2), dim = 1))
        PET_downsample2_up = self.conv_upsample3_PET(l3_output_PET)
        l2_output_PET = self.decoder2_PET(torch.cat((PET_downsample2_up, PET_downsample1), dim = 1))
        PET_downsample1_up = self.conv_upsample4_PET(l2_output_PET)
        l1_output_PET = self.decoder1_PET(torch.cat((PET_downsample1_up, PET), dim = 1))
        spatial_attention_map = self.classifier_PET(l1_output_PET)

        spatial_attention_map_div2 = F.interpolate(spatial_attention_map, scale_factor = 0.5, mode ='trilinear')
        spatial_attention_map_div4 = F.interpolate(spatial_attention_map, scale_factor = 0.25, mode ='trilinear')
        spatial_attention_map_div8 = F.interpolate(spatial_attention_map, scale_factor = 0.125, mode ='trilinear')

        CT = self.expand_channels_CT(CT)
        CT_downsample1 = self.conv_downsample1_CT(CT)
        CT_downsample2 = self.conv_downsample2_CT(CT_downsample1)
        CT_downsample3 = self.conv_downsample3_CT(CT_downsample2)
        CT_downsample4 = self.conv_downsample4_CT(CT_downsample3)

        CT_downsample4_up = self.conv_upsample1_CT(CT_downsample4)
        l4_output_CT = self.decoder4_CT(torch.cat((CT_downsample4_up, CT_downsample3*spatial_attention_map_div8), dim = 1))
        CT_downsample3_up = self.conv_upsample2_CT(l4_output_CT)
        l3_output_CT = self.decoder3_CT(torch.cat((CT_downsample3_up, CT_downsample2*spatial_attention_map_div4), dim = 1))
        CT_downsample2_up = self.conv_upsample3_CT(l3_output_CT)
        l2_output_CT = self.decoder2_CT(torch.cat((CT_downsample2_up, CT_downsample1*spatial_attention_map_div2), dim = 1))
        CT_downsample1_up = self.conv_upsample4_CT(l2_output_CT)
        l1_output_CT = self.decoder1_CT(torch.cat((CT_downsample1_up, CT*spatial_attention_map), dim = 1))

        l1_output = self.classifier1(l1_output_CT)
        l2_output = self.upsample2(self.classifier2(l2_output_CT))
        l3_output = self.upsample3(self.classifier3(l3_output_CT))
        l4_output = self.upsample4(self.classifier4(l4_output_CT))
        
        return [l1_output, l2_output, l3_output, l4_output]

    
class UNet3d_MultiModalV2(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 32, channel_bound = 320, vectorization = True):
        super().__init__()
        self.vectorization = vectorization
        channel_bound = min(n_channels * 16, channel_bound)
        self.expand_channels_CT = aspp_encoder(in_channels, n_channels, stride = 1)
        self.conv_downsample1_CT = encoder_block(n_channels, n_channels * 2)
        self.conv_downsample2_CT = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3_CT = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4_CT = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1_CT = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2_CT= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3_CT= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4_CT= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1_CT = encoder_block(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2_CT = encoder_block(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3_CT = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4_CT = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)

        self.expand_channels_PET = aspp_encoder(in_channels, n_channels, stride = 1)
        self.conv_downsample1_PET = encoder_block(n_channels, n_channels * 2)
        self.conv_downsample2_PET = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3_PET = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4_PET = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1_PET = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2_PET= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3_PET= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4_PET= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1_PET = encoder_block(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2_PET = encoder_block(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3_PET = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4_PET = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv3d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv3d(n_channels * 2 * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv3d(n_channels * 4 * 2, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv3d(n_channels * 8 * 2, out_channels, kernel_size = 1)


        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)


    def forward(self, CT, PET):
        PET = self.expand_channels_PET(PET)
        PET_downsample1 = self.conv_downsample1_PET(PET)
        PET_downsample2 = self.conv_downsample2_PET(PET_downsample1)
        PET_downsample3 = self.conv_downsample3_PET(PET_downsample2)
        PET_downsample4 = self.conv_downsample4_PET(PET_downsample3)

        PET_downsample4_up = self.conv_upsample1_PET(PET_downsample4)
        l4_output_PET = self.decoder4_PET(torch.cat((PET_downsample4_up, PET_downsample3), dim = 1))
        PET_downsample3_up = self.conv_upsample2_PET(l4_output_PET)
        l3_output_PET = self.decoder3_PET(torch.cat((PET_downsample3_up, PET_downsample2), dim = 1))
        PET_downsample2_up = self.conv_upsample3_PET(l3_output_PET)
        l2_output_PET = self.decoder2_PET(torch.cat((PET_downsample2_up, PET_downsample1), dim = 1))
        PET_downsample1_up = self.conv_upsample4_PET(l2_output_PET)
        l1_output_PET = self.decoder1_PET(torch.cat((PET_downsample1_up, PET), dim = 1))
        CT = self.expand_channels_CT(CT)
        CT_downsample1 = self.conv_downsample1_CT(CT)
        CT_downsample2 = self.conv_downsample2_CT(CT_downsample1)
        CT_downsample3 = self.conv_downsample3_CT(CT_downsample2)
        CT_downsample4 = self.conv_downsample4_CT(CT_downsample3)

        CT_downsample4_up = self.conv_upsample1_CT(CT_downsample4)
        l4_output_CT = self.decoder4_CT(torch.cat((CT_downsample4_up, CT_downsample3), dim = 1))
        CT_downsample3_up = self.conv_upsample2_CT(l4_output_CT)
        l3_output_CT = self.decoder3_CT(torch.cat((CT_downsample3_up, CT_downsample2), dim = 1))
        CT_downsample2_up = self.conv_upsample3_CT(l3_output_CT)
        l2_output_CT = self.decoder2_CT(torch.cat((CT_downsample2_up, CT_downsample1), dim = 1))
        CT_downsample1_up = self.conv_upsample4_CT(l2_output_CT)
        l1_output_CT = self.decoder1_CT(torch.cat((CT_downsample1_up, CT), dim = 1))

        l1_output = self.classifier1(torch.cat((l1_output_CT, l1_output_PET), dim = 1))
        l2_output = self.upsample2(self.classifier2(torch.cat((l2_output_CT, l2_output_PET), dim = 1)))
        l3_output = self.upsample3(self.classifier3(torch.cat((l3_output_CT, l3_output_PET), dim = 1)))
        l4_output = self.upsample4(self.classifier4(torch.cat((l4_output_CT, l4_output_PET), dim = 1)))

        if self.vectorization:
            l1_output = torch.softmax(l1_output , dim = 1)
            l2_output = torch.softmax(l2_output , dim = 1)
            l3_output = torch.softmax(l3_output , dim = 1)
            l4_output = torch.softmax(l4_output , dim = 1)
            result = torch.stack([l1_output, l2_output, l3_output, l4_output]).permute(1,0,2,3,4,5)

            return result.contiguous()

        
        return [l1_output, l2_output, l3_output, l4_output]


class UNet3d_Plus_ASPP(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 32, channel_bound = 320):
        super().__init__()
        self.expand_channels = aspp_encoder(in_channels, n_channels, stride = 1)
        self.conv_downsample1 = aspp_encoder(n_channels, n_channels * 2)
        self.conv_downsample2 = encoder_block(n_channels * 2, n_channels * 4)
        self.conv_downsample3 = encoder_block(n_channels * 4, n_channels * 8)
        self.conv_downsample4 = encoder_block(n_channels * 8, channel_bound)

        self.conv_upsample1 = nn.ConvTranspose3d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2= nn.ConvTranspose3d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3= nn.ConvTranspose3d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4= nn.ConvTranspose3d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1 = aspp_encoder(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2 = aspp_encoder(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3 = encoder_block(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4 = encoder_block(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv3d(n_channels, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv3d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv3d(n_channels * 4, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv3d(n_channels * 8, out_channels, kernel_size = 1)

        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)

    
    def forward(self, x):
        x = self.expand_channels(x)
        x_downsample1 = self.conv_downsample1(x)
        x_downsample2 = self.conv_downsample2(x_downsample1)
        x_downsample3 = self.conv_downsample3(x_downsample2)
        x_downsample4 = self.conv_downsample4(x_downsample3)

        x_downsample4_up = self.conv_upsample1(x_downsample4)
        l4_output = self.decoder4(torch.cat((x_downsample4_up, x_downsample3), dim = 1))
        x_downsample3_up = self.conv_upsample2(l4_output)
        l3_output = self.decoder3(torch.cat((x_downsample3_up, x_downsample2), dim = 1))
        x_downsample2_up = self.conv_upsample3(l3_output)
        l2_output = self.decoder2(torch.cat((x_downsample2_up, x_downsample1), dim = 1))
        x_downsample1_up = self.conv_upsample4(l2_output)
        l1_output = self.decoder1(torch.cat((x_downsample1_up, x), dim = 1))

        l1_output = self.classifier1(l1_output)
        l2_output = self.upsample2(self.classifier2(l2_output))
        l3_output = self.upsample3(self.classifier3(l3_output))
        l4_output = self.upsample4(self.classifier4(l4_output))

        
        return [l1_output, l2_output, l3_output, l4_output]



class UNet2d_Plus(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 64, channel_bound = 1024):
        super().__init__()
        self.expand_channels = encoder_block_2d(in_channels, n_channels, stride = 1)
        self.conv_downsample1 = encoder_block_2d(n_channels, n_channels * 2)
        self.conv_downsample2 = encoder_block_2d(n_channels * 2, n_channels * 4)
        self.conv_downsample3 = encoder_block_2d(n_channels * 4, n_channels * 8)
        self.conv_downsample4 = encoder_block_2d(n_channels * 8, channel_bound)

        self.conv_upsample1 = nn.ConvTranspose2d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2= nn.ConvTranspose2d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3= nn.ConvTranspose2d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4= nn.ConvTranspose2d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1 = encoder_block_2d(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2 = encoder_block_2d(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3 = encoder_block_2d(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4 = encoder_block_2d(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv2d(n_channels, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv2d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv2d(n_channels * 4, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv2d(n_channels * 8, out_channels, kernel_size = 1)

        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)

    
    def forward(self, x):
        x = self.expand_channels(x)
        x_downsample1 = self.conv_downsample1(x)
        x_downsample2 = self.conv_downsample2(x_downsample1)
        x_downsample3 = self.conv_downsample3(x_downsample2)
        x_downsample4 = self.conv_downsample4(x_downsample3)

        x_downsample4_up = self.conv_upsample1(x_downsample4)
        l4_output = self.decoder4(torch.cat((x_downsample4_up, x_downsample3), dim = 1))
        x_downsample3_up = self.conv_upsample2(l4_output)
        l3_output = self.decoder3(torch.cat((x_downsample3_up, x_downsample2), dim = 1))
        x_downsample2_up = self.conv_upsample3(l3_output)
        l2_output = self.decoder2(torch.cat((x_downsample2_up, x_downsample1), dim = 1))
        x_downsample1_up = self.conv_upsample4(l2_output)
        l1_output = self.decoder1(torch.cat((x_downsample1_up, x), dim = 1))

        l1_output = self.classifier1(l1_output)
        l2_output = self.upsample2(self.classifier2(l2_output))
        l3_output = self.upsample3(self.classifier3(l3_output))
        l4_output = self.upsample4(self.classifier4(l4_output))

        
        return [l1_output, l2_output, l3_output, l4_output]




class UNet2d_GRU(nn.Module):
    def __init__(self, in_channels = 1, out_channels = 2, n_channels = 64, channel_bound = 1024, input_size = (48, 48), device = 0):
        super().__init__()
        input_size = np.array(input_size)
        self.device = device
        self.expand_channels = encoder_block_2d(in_channels, n_channels, stride = 1)
        self.conv_downsample1 = encoder_block_2d(n_channels, n_channels * 2)
        self.conv_downsample2 = encoder_block_2d(n_channels * 2, n_channels * 4)
        self.conv_downsample3 = encoder_block_2d(n_channels * 4, n_channels * 8)
        self.conv_downsample4 = encoder_block_2d(n_channels * 8, channel_bound)

        self.conv_upsample1 = nn.ConvTranspose2d(channel_bound, n_channels * 8, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample2= nn.ConvTranspose2d(n_channels * 8, n_channels * 4, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample3= nn.ConvTranspose2d(n_channels * 4, n_channels * 2, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)
        self.conv_upsample4= nn.ConvTranspose2d(n_channels * 2, n_channels, kernel_size = 3, stride = 2, padding = 1, output_padding = 1)

        self.decoder1 = encoder_block_2d(n_channels * 1 * 2, n_channels, stride = 1)
        self.decoder2 = encoder_block_2d(n_channels * 2 * 2, n_channels * 2, stride = 1)
        self.decoder3 = encoder_block_2d(n_channels * 4 * 2, n_channels * 4, stride = 1)
        self.decoder4 = encoder_block_2d(n_channels * 8 * 2, n_channels * 8, stride = 1)
        
        self.classifier1 = nn.Conv2d(n_channels, out_channels, kernel_size = 1)
        self.classifier2 = nn.Conv2d(n_channels * 2, out_channels, kernel_size = 1)
        self.classifier3 = nn.Conv2d(n_channels * 4, out_channels, kernel_size = 1)
        self.classifier4 = nn.Conv2d(n_channels * 8, out_channels, kernel_size = 1)

        self.upsample4 = nn.Upsample(scale_factor=8)
        self.upsample3 = nn.Upsample(scale_factor=4)
        self.upsample2 = nn.Upsample(scale_factor=2)

        self.gru1 = ConvGRUCell(input_size = input_size, input_dim = n_channels, hidden_dim = n_channels, kernel_size = [3, 3])
        self.gru2 = ConvGRUCell(input_size = input_size //2, input_dim = n_channels*2, hidden_dim = n_channels*2, kernel_size = [3, 3])
        self.gru3 = ConvGRUCell(input_size = input_size //4, input_dim = n_channels*4, hidden_dim = n_channels*4, kernel_size = [3, 3])
        self.gru4 = ConvGRUCell(input_size = input_size //8, input_dim = n_channels*8, hidden_dim = n_channels*8, kernel_size = [3, 3])
        self.gru5 = ConvGRUCell(input_size = input_size //16, input_dim = n_channels*16, hidden_dim = n_channels*16, kernel_size = [3, 3])

    
    def forward(self, x, hidden_state = None):
        x = self.expand_channels(x)
        x_downsample1 = self.conv_downsample1(x)
        x_downsample2 = self.conv_downsample2(x_downsample1)
        x_downsample3 = self.conv_downsample3(x_downsample2)
        x_downsample4 = self.conv_downsample4(x_downsample3)

        if hidden_state != None:
            x = self.gru1(x, hidden_state[0])
            x_downsample1 = self.gru2(x_downsample1, hidden_state[1])
            x_downsample2 = self.gru3(x_downsample2, hidden_state[2])
            x_downsample3 = self.gru4(x_downsample3, hidden_state[3])
            x_downsample4 = self.gru5(x_downsample4, hidden_state[4])
        else:
            x = self.gru1(x, torch.zeros(x.shape).float().to(self.device))
            x_downsample1 = self.gru2(x_downsample1, torch.zeros(x_downsample1.shape).float().to(self.device))
            x_downsample2 = self.gru3(x_downsample2, torch.zeros(x_downsample2.shape).float().to(self.device))
            x_downsample3 = self.gru4(x_downsample3, torch.zeros(x_downsample3.shape).float().to(self.device))
            x_downsample4 = self.gru5(x_downsample4, torch.zeros(x_downsample4.shape).float().to(self.device))

        x_downsample4_up = self.conv_upsample1(x_downsample4)
        l4_output = self.decoder4(torch.cat((x_downsample4_up, x_downsample3), dim = 1))
        x_downsample3_up = self.conv_upsample2(l4_output)
        l3_output = self.decoder3(torch.cat((x_downsample3_up, x_downsample2), dim = 1))
        x_downsample2_up = self.conv_upsample3(l3_output)
        l2_output = self.decoder2(torch.cat((x_downsample2_up, x_downsample1), dim = 1))
        x_downsample1_up = self.conv_upsample4(l2_output)
        l1_output = self.decoder1(torch.cat((x_downsample1_up, x), dim = 1))

        l1_output = self.classifier1(l1_output)
        l2_output = self.upsample2(self.classifier2(l2_output))
        l3_output = self.upsample3(self.classifier3(l3_output))
        l4_output = self.upsample4(self.classifier4(l4_output))

        
        return [l1_output, l2_output, l3_output, l4_output], [x.detach().clone(), x_downsample1.detach().clone(), x_downsample2.detach().clone(), x_downsample3.detach().clone(), x_downsample4.detach().clone()]


